@@ -11,13 +11,37 @@ from . import utils
 
 
 class DatasetFactory():
+    """Base class for creating datasets.
+
+    Attributes:
+      df (pd.DataFrame): A full dataframe of the data.
+      df_ml (pd.DataFrame): A dataframe of data for machine learning models.
+      download (module): Script for downloading the dataset.
+      metadata (dict): Metadata of the dataset.
+      root (str): Root directory for the data.
+      unknown_name (str): Name of the unknown class.
+    """
+
+    unknown_name = 'unknown'
+
     def __init__(
-        self, 
-        root: str,
-        df: Optional[pd.DataFrame] = None,
-        download: bool = False,
-        **kwargs
-        ):
+            self, 
+            root: str,
+            df: Optional[pd.DataFrame] = None,
+            download: bool = False,
+            **kwargs) -> None:
+        """Initializes the class.
+
+        If `download`, then it downloads the data.
+        If `df` is specified, it copies it. Otherwise, it creates it
+        by the `create_catalogue` method.
+        It creates `df_ml` by the `create_catalogue_ml` method.
+
+        Args:
+            root (str): Root directory for the data.
+            df (Optional[pd.DataFrame], optional): A full dataframe of the data.
+            download (bool, optional): Whether to download the data.
+        """
 
         self.root = root
         if download and hasattr(self, 'download'): 
@@ -26,18 +50,52 @@ class DatasetFactory():
             self.df = self.create_catalogue(**kwargs)
         else:
             self.df = df
+        self.df_ml = self.create_catalogue_ml(self.df)
 
     def create_catalogue(self):
-        '''
-        Creates a dataframe catalogue summarizing the dataset.
-        This method is dataset specific and each dataset needs to override it.
-        '''
+        """Creates the dataframe.
+
+        Raises:
+            NotImplementedError: Needs to be implemented by subclasses.
+        """
+
         raise NotImplementedError()
 
+    def create_catalogue_ml(self, df: pd.DataFrame) -> pd.DataFrame:
+        """Creates the dataframe for machine learning models.
+
+        It removes classes with only one sample.
+        It removes unknown identities.
+        It reset the dataframe index.
+
+        Args:
+            df (pd.DataFrame): A full dataframe of the data.
+
+        Returns:
+            A dataframe of data for machine learning models.
+        """
+
+        df = df.groupby('identity').filter(lambda x : len(x) >= 2)
+        df = df[df['identity'] != self.unknown_name]
+        df.reset_index(drop=True, inplace=True)
+        return df
+    
     def finalize_catalogue(self, df: pd.DataFrame) -> pd.DataFrame:
-        '''
-        Finalizes catalogue dataframe and runs checks for errors.
-        '''
+        """Reorders the dataframe and check file paths.
+
+        Reorders the columns and removes constant columns.
+        Checks if columns are in correct formats.
+        Checks if ids are unique and if all files exist.
+
+        Args:
+            df (pd.DataFrame): A full dataframe of the data.
+
+        Returns:
+            A full dataframe of the data, slightly modified.
+        """
+
+        self.check_required_columns(df)
+        self.check_types_columns(df)
         df = self.reorder_df(df)
         df = self.remove_constant_columns(df)
         self.check_unique_id(df)
@@ -46,12 +104,104 @@ class DatasetFactory():
             self.check_files_exist(df['segmentation'])
         return df
 
+    def check_required_columns(self, df: pd.DataFrame) -> None:
+        """Check if all required columns are present.
+
+        Args:
+            df (pd.DataFrame): A full dataframe of the data.
+        """
+
+        for col_name in ['id', 'identity', 'path']:
+            if col_name not in df.columns:
+                raise(Exception('Column %s must be in the dataframe columns.' % col_name))
+
+    def check_types_columns(self, df: pd.DataFrame) -> None:
+        """Checks if columns are in correct formats.
+
+        The format are specified in `requirements`, which is list
+        of tuples. The first value is the name of the column
+        and the second value is a list of formats. The column
+        must be at least one of the formats.
+
+        Args:
+            df (pd.DataFrame): A full dataframe of the data.
+        """
+
+        requirements = [
+            ('id', ['int', 'str']),
+            ('identity', ['int', 'str']),
+            ('path', ['str']),
+            ('bbox', ['list_numeric']),
+            ('date', ['date']),
+            ('keypoints', ['list_numeric']),
+            ('position', ['str']),
+            ('species', ['str', 'list']),
+            ('video', ['int']),
+        ]
+        # Verify if the columns are in correct formats
+        for col_name, allowed_types in requirements:
+            if col_name in df.columns:
+                # Remove empty values to be sure
+                col = df[col_name][~df[col_name].isnull()]
+                if len(col) > 0:
+                    self.check_types_column(col, col_name, allowed_types)
+    
+    def check_types_column(self, col: pd.Series, col_name: str, allowed_types: List[str]) -> None:
+        """Checks if the column `col` is in the format `allowed_types`.
+
+        Args:
+            col (pd.Series): Column to be checked.
+            col_name (str): Column name used only for raising exceptions.
+            allowed_types (List[str]): List of strings with allowed values:
+                `int` (all values must be integers),
+                `str` (strings),
+                `list` (lists),
+                `list_numeric` (lists with numeric values),
+                `date` (dates as tested by `pd.to_datetime`).
+        """
+
+        if 'int' in allowed_types and pd.api.types.is_integer_dtype(col):
+            return None
+        if 'str' in allowed_types and pd.api.types.is_string_dtype(col):
+            return None
+        if 'list' in allowed_types and pd.api.types.is_list_like(col):
+            check = True
+            for val in col:
+                if not pd.api.types.is_list_like(val):
+                    check = False
+                    break
+            if check:                
+                return None        
+        if 'list_numeric' in allowed_types and pd.api.types.is_list_like(col):
+            check = True
+            for val in col:            
+                if not pd.api.types.is_list_like(val) and not pd.api.types.is_numeric_dtype(pd.Series(val)):
+                    check = False
+                    break
+            if check:                
+                return None
+        if 'date' in allowed_types:
+            try:
+                pd.to_datetime(col)
+                return None
+            except:
+                pass
+        raise(Exception('Column %s has wrong type. Allowed types = %s' % (col_name, str(allowed_types))))
+
     def reorder_df(self, df: pd.DataFrame) -> pd.DataFrame:
-        '''
-        Reorders columns in the dataframe.
-        Columns specified in default_order go first.
-        '''
-        default_order = ['id', 'identity', 'path', 'bbox', 'segmentation', 'position', 'species', 'keypoints']
+        """Reorders rows and columns in the dataframe.
+
+        Rows are sorted based on id.
+        Columns are reorder based on the `default_order` list.
+
+        Args:
+            df (pd.DataFrame): A full dataframe of the data.
+
+        Returns:
+            A full dataframe of the data, slightly modified.
+        """
+
+        default_order = ['id', 'identity', 'path', 'bbox', 'date', 'keypoints', 'orientation', 'segmentation', 'species']
         df_names = list(df.columns)
         col_names = []
         for name in default_order:
@@ -65,48 +215,41 @@ class DatasetFactory():
         return df.reindex(columns=col_names)
 
     def remove_constant_columns(self, df: pd.DataFrame) -> pd.DataFrame:
-        '''
-        Removes columns with single unique value.
-        '''
+        """Removes columns with a single unique value.
+
+        Args:
+            df (pd.DataFrame): A full dataframe of the data.
+
+        Returns:
+            A full dataframe of the data, slightly modified.
+        """ 
+
         for df_name in list(df.columns):
             if df[df_name].astype('str').nunique() == 1:
                 df = df.drop([df_name], axis=1)
         return df
 
     def check_unique_id(self, df: pd.DataFrame) -> None:
-        '''
-        Checks if values in the 'id' column are unique.
-        '''
+        """Checks if values in the id column are unique.
+
+        Args:
+            df (pd.DataFrame): A full dataframe of the data.
+        """
+
         if len(df['id'].unique()) != len(df):
             raise(Exception('Image ID not unique.'))
 
     def check_files_exist(self, col: pd.Series) -> None:
-        '''
-        Checks if paths in a given column exist.
-        '''
+        """Checks if paths in a given column exist.
+
+        Args:
+            col (pd.Series): A column of a dataframe.
+        """
+
         for path in col:
             if type(path) == str and not os.path.exists(os.path.join(self.root, path)):
                 raise(Exception('Path does not exist:' + os.path.join(self.root, path)))
 
-    def split_data(self, splitter):
-        '''
-        Splits data by a splitter class with scikit-like API.
-        '''
-        indexes = np.arange(len(self.df))
-        labels = self.df['identity']
-
-        splits = []
-        for train, valid in splitter.split(indexes, labels):
-            splits.append([
-                self.df.iloc[indexes[train]],
-                self.df.iloc[indexes[valid]],
-            ])
-
-        if len(splits) == 1:
-            return splits[0]
-        else:
-            return splits
-
 
 class DatasetFactoryWildMe(DatasetFactory):
     def create_catalogue_wildme(self, prefix: str, year: int) -> pd.DataFrame:
@@ -124,7 +267,7 @@ class DatasetFactoryWildMe(DatasetFactory):
                 raise(Exception('Wrong number of segmentations'))
         
         # Extract the data from the JSON file
-        create_dict = lambda i: {'identity': i['name'], 'bbox': utils.segmentation_bbox(i['segmentation'][0]), 'image_id': i['image_id'], 'category_id': i['category_id'], 'segmentation': i['segmentation'][0], 'position': i['viewpoint']}
+        create_dict = lambda i: {'identity': i['name'], 'bbox': utils.segmentation_bbox(i['segmentation'][0]), 'image_id': i['image_id'], 'category_id': i['category_id'], 'segmentation': i['segmentation'][0], 'orientation': i['viewpoint']}
         df_annotation = pd.DataFrame([create_dict(i) for i in data['annotations']])
         create_dict = lambda i: {'file_name': i['file_name'], 'image_id': i['id'], 'date': i['date_captured']}
         df_images = pd.DataFrame([create_dict(i) for i in data['images']])
@@ -138,7 +281,7 @@ class DatasetFactoryWildMe(DatasetFactory):
         # Modify some columns
         df['path'] = path_images + os.path.sep + df['file_name']
         df['id'] = range(len(df))    
-        df.loc[df['identity'] == '____', 'identity'] = 'unknown'
+        df.loc[df['identity'] == '____', 'identity'] = self.unknown_name
 
         # Remove segmentations which are the same as bounding boxes
         ii = []
@@ -182,9 +325,9 @@ class AAUZebraFishID(DatasetFactory):
         attributes = attributes.astype(int).astype(bool)
 
         # Split additional data into a separate structure
-        position = data['Right,Turning,Occlusion,Glitch'].str.split(',', expand=True)[0]
-        position.replace('1', 'right', inplace=True)
-        position.replace('0', 'left', inplace=True)
+        orientation = data['Right,Turning,Occlusion,Glitch'].str.split(',', expand=True)[0]
+        orientation.replace('1', 'right', inplace=True)
+        orientation.replace('0', 'left', inplace=True)
 
         # Modify information about video sources
         video = data['Filename'].str.split('_',  expand=True)[0]
@@ -197,7 +340,7 @@ class AAUZebraFishID(DatasetFactory):
             'identity': data['Object ID'],
             'video': video,
             'bbox': bbox,
-            'position': position,
+            'orientation': orientation,
         })
         df = df.join(attributes)
         return self.finalize_catalogue(df)
@@ -217,7 +360,7 @@ class AerialCattle2017(DatasetFactory):
             'id': utils.create_id(data['file']),
             'path': data['path'] + os.path.sep + data['file'],
             'identity': folders[1].astype(int),
-            'video': folders[2],
+            'video': folders[2].astype(int),
         })
         return self.finalize_catalogue(df)
 
@@ -354,7 +497,7 @@ class BirdIndividualID(DatasetFactory):
         df2 = pd.DataFrame({    
             'id': utils.create_id(data['file']),
             'path': self.prefix1 + os.path.sep + 'New_birds' + os.path.sep + data['path'] + os.path.sep + data['file'],
-            'identity': 'unknown',
+            'identity': self.unknown_name,
             'species': species,
             'split': 'unassigned',
         })
@@ -362,6 +505,7 @@ class BirdIndividualID(DatasetFactory):
 
 
 class BirdIndividualIDSegmented(BirdIndividualID):
+    download = downloads.segmented
     prefix1 = 'Cropped_pictures'
     prefix2 = 'IndividuaID'
 
@@ -373,7 +517,7 @@ class CTai(DatasetFactory):
     def create_catalogue(self) -> pd.DataFrame:
         # Define the wrong identity names
         replace_names = [
-            ('Adult', 'unknown'),
+            ('Adult', self.unknown_name),
             ('Akouba', 'Akrouba'),
             ('Freddy', 'Fredy'),
             ('Ibrahiim', 'Ibrahim'),
@@ -455,8 +599,20 @@ class Cows2021(DatasetFactory):
             'path': data['path'] + os.path.sep + data['file'],
             'identity': folders[4].astype(int),
         })
+        df['date'] = df['path'].apply(lambda x: self.extract_date(x))
         return self.finalize_catalogue(df)
 
+    def extract_date(self, x):
+        x = os.path.split(x)[1]
+        if x.startswith('image_'):
+            x = x[6:]
+        if x[7] == '_':
+            x = x[8:]
+        i1 = x.find('_')
+        i2 = x[i1+1:].find('_')
+        x = x[:i1+i2+1]
+        return datetime.datetime.strptime(x, '%Y-%m-%d_%H-%M-%S').strftime('%Y-%m-%d %H:%M:%S')
+
 
 class Drosophila(DatasetFactory):
     download = downloads.drosophila
@@ -468,7 +624,7 @@ class Drosophila(DatasetFactory):
         folders = data['path'].str.split(os.path.sep, expand=True)
 
         # Extract information from the folder structure
-        data['identity'] = 'unknown'
+        data['identity'] = self.unknown_name
         for i_week in range(1, 4):
             idx1 = folders[0].str.startswith('week' + str(i_week))
             idx2 = folders[1] == 'val'
@@ -592,7 +748,7 @@ class HappyWhale(DatasetFactory):
         df2 = pd.DataFrame({
             'id': test_files.str.split('.', expand=True)[0],
             'path': 'test_images' + os.path.sep + test_files,
-            'identity': 'unknown',
+            'identity': self.unknown_name,
             'species': np.nan,
             'split': 'test'
             })
@@ -609,7 +765,7 @@ class HumpbackWhaleID(DatasetFactory):
     def create_catalogue(self) -> pd.DataFrame:
         # Load the training data
         data = pd.read_csv(os.path.join(self.root, 'train.csv'))
-        data.loc[data['Id'] == 'new_whale', 'Id'] = 'unknown'
+        data.loc[data['Id'] == 'new_whale', 'Id'] = self.unknown_name
         df1 = pd.DataFrame({
             'id': data['Image'].str.split('.', expand=True)[0],
             'path': 'train' + os.path.sep + data['Image'],
@@ -626,7 +782,7 @@ class HumpbackWhaleID(DatasetFactory):
         df2 = pd.DataFrame({
             'id': test_files.str.split('.', expand=True)[0],
             'path': 'test' + os.path.sep + test_files,
-            'identity': 'unknown',
+            'identity': self.unknown_name,
             'split': 'test'
             })
         
@@ -744,7 +900,7 @@ class NDD20(DatasetFactory):
                 if 'id' in region['region_attributes']:
                     identity = region['region_attributes']['id']
                 else:
-                    identity = 'unknown'
+                    identity = self.unknown_name
                 segmentation = np.zeros(2*len(region['shape_attributes']['all_points_x']))
                 segmentation[0::2] = region['shape_attributes']['all_points_x']
                 segmentation[1::2] = region['shape_attributes']['all_points_y']
@@ -755,7 +911,7 @@ class NDD20(DatasetFactory):
                     'file_name': data[key]['filename'],
                     'reg_type': region['shape_attributes']['name'],
                     'segmentation': segmentation,
-                    'position': 'above'
+                    'orientation': 'above'
                 })
         
         # Load information about the below-water dataset
@@ -769,7 +925,7 @@ class NDD20(DatasetFactory):
                 if 'id' in region['region_attributes']:
                     identity = region['region_attributes']['id']
                 else:
-                    identity = 'unknown'
+                    identity = self.unknown_name
                 segmentation = np.zeros(2*len(region['shape_attributes']['all_points_x']))
                 segmentation[0::2] = region['shape_attributes']['all_points_x']
                 segmentation[1::2] = region['shape_attributes']['all_points_y']
@@ -780,7 +936,7 @@ class NDD20(DatasetFactory):
                     'file_name': data[key]['filename'],
                     'reg_type': region['shape_attributes']['name'],
                     'segmentation': segmentation,
-                    'position': 'below'
+                    'orientation': 'below'
                 })
 
         # Create the dataframe from entries 
@@ -790,7 +946,7 @@ class NDD20(DatasetFactory):
 
         # Finalize the dataframe
         df['id'] = range(len(df))
-        df['path'] = df['position'].str.upper() + os.path.sep + df['file_name']
+        df['path'] = df['orientation'].str.upper() + os.path.sep + df['file_name']
         df = df.drop(['reg_type', 'file_name'], axis=1)
         return self.finalize_catalogue(df)
 
@@ -814,7 +970,7 @@ class NOAARightWhale(DatasetFactory):
         df2 = pd.DataFrame({
             'id': data['Image'].str.split('.', expand=True)[0].str.strip('w_').astype(int),
             'path': 'imgs' + os.path.sep + data['Image'],
-            'identity': 'unknown',
+            'identity': self.unknown_name,
             })
         
         # Finalize the dataframe
@@ -831,18 +987,18 @@ class NyalaData(DatasetFactory):
         data = utils.find_images(self.root)
         folders = data['path'].str.split(os.path.sep, expand=True)
 
-        # Extract information from the folder structure and about position
+        # Extract information from the folder structure and about orientation
         identity = folders[3].astype(int)
-        position = np.full(len(data), np.nan, dtype=object)
-        position[['left' in filename for filename in data['file']]] = 'left'
-        position[['right' in filename for filename in data['file']]] = 'right'
+        orientation = np.full(len(data), np.nan, dtype=object)
+        orientation[['left' in filename for filename in data['file']]] = 'left'
+        orientation[['right' in filename for filename in data['file']]] = 'right'
 
         # Finalize the dataframe
         df = pd.DataFrame({
             'id': utils.create_id(data['file']),
             'path': data['path'] + os.path.sep + data['file'],
             'identity': identity,
-            'position': position,
+            'orientation': orientation,
         })
         return self.finalize_catalogue(df)   
 
@@ -895,6 +1051,7 @@ class SealID(DatasetFactory):
 
 
 class SealIDSegmented(SealID):
+    download = downloads.segmented
     prefix = 'segmented_'
 
 
@@ -904,26 +1061,52 @@ class SeaTurtleID(DatasetFactory):
 
     def create_catalogue(self) -> pd.DataFrame:
         # Load annotations JSON file
-        path_json = 'annotations.json'
+        path_json = 'annotations_new.json'
         with open(os.path.join(self.root, path_json)) as file:
             data = json.load(file)
 
         # Extract dtaa from the JSON file
-        create_dict = lambda i: {'id': i['id'], 'bbox': i['bbox'], 'image_id': i['image_id'], 'identity': i['identity'], 'segmentation': i['segmentation'], 'position': i['position']}
+        create_dict = lambda i: {'id': i['id'], 'bbox': i['bbox'], 'image_id': i['image_id'], 'identity': i['identity'], 'segmentation': i['segmentation'], 'orientation': i['position']}
         df_annotation = pd.DataFrame([create_dict(i) for i in data['annotations']])
-        create_dict = lambda i: {'file_name': i['path'].split('/')[-1], 'image_id': i['id'], 'date': i['date'], 'path_orig': i['path_orig']}
+        create_dict = lambda i: {'file_name': i['path'].split('/')[-1], 'image_id': i['id'], 'date': i['date']}
         df_images = pd.DataFrame([create_dict(i) for i in data['images']])
 
         # Merge the information from the JSON file
         df = pd.merge(df_annotation, df_images, on='image_id')
         df['path'] = 'images' + os.path.sep + df['identity'] + os.path.sep + df['file_name']        
-        df = df.drop(['image_id', 'file_name', 'path_orig'], axis=1)
+        df = df.drop(['image_id', 'file_name'], axis=1)
         df['date'] = df['date'].apply(lambda x: x[:4] + '-' + x[5:7] + '-' + x[8:])
 
         # Finalize the dataframe
         return self.finalize_catalogue(df)
-    
-        
+
+
+class SeaTurtleIDHeads(DatasetFactory):
+    download = downloads.sea_turtle_id_heads
+    metadata = metadata['SeaTurtleIDHeads']
+
+    def create_catalogue(self) -> pd.DataFrame:
+        # Load annotations JSON file
+        path_json = 'annotations.json'
+        with open(os.path.join(self.root, path_json)) as file:
+            data = json.load(file)
+
+        # Extract dtaa from the JSON file
+        create_dict = lambda i: {'id': i['id'], 'image_id': i['image_id'], 'identity': i['identity'], 'orientation': i['position']}
+        df_annotation = pd.DataFrame([create_dict(i) for i in data['annotations']])
+        create_dict = lambda i: {'file_name': i['path'].split('/')[-1], 'image_id': i['id'], 'date': i['date']}
+        df_images = pd.DataFrame([create_dict(i) for i in data['images']])
+
+        # Merge the information from the JSON file
+        df = pd.merge(df_annotation, df_images, on='image_id')
+        df['path'] = 'images' + os.path.sep + df['identity'] + os.path.sep + df['file_name']        
+        df = df.drop(['image_id', 'file_name'], axis=1)
+        df['date'] = df['date'].apply(lambda x: x[:4] + '-' + x[5:7] + '-' + x[8:])
+
+        # Finalize the dataframe
+        return self.finalize_catalogue(df)
+
+
 class SMALST(DatasetFactory):
     download = downloads.smalst
     metadata = metadata['SMALST']
@@ -969,7 +1152,7 @@ class StripeSpotter(DatasetFactory):
         # Load additional information
         data_aux = pd.read_csv(os.path.join(self.root, 'data', 'SightingData.csv'))
         data = pd.merge(data, data_aux, how='left', left_on='index', right_on='#imgindex')
-        data.loc[data['animal_name'].isnull(), 'animal_name'] = 'unknown'
+        data.loc[data['animal_name'].isnull(), 'animal_name'] = self.unknown_name
         
         # Finalize the dataframe
         df = pd.DataFrame({
@@ -977,7 +1160,7 @@ class StripeSpotter(DatasetFactory):
             'path':  data['path'] + os.path.sep + data['file'],
             'identity': data['animal_name'],
             'bbox': pd.Series([[int(a) for a in b.split(' ')] for b in data['roi']]),
-            'position': data['flank'],
+            'orientation': data['flank'],
             'photo_quality': data['photo_quality'],
         })
         return self.finalize_catalogue(df)  
@@ -1049,12 +1232,12 @@ class ZindiTurtleRecall(DatasetFactory):
         # Finalize the dataframe
         data = pd.concat([data_train, data_test, data_extra])
         data = data.reset_index(drop=True)        
-        data.loc[data['turtle_id'].isnull(), 'turtle_id'] = 'unknown'
+        data.loc[data['turtle_id'].isnull(), 'turtle_id'] = self.unknown_name
         df = pd.DataFrame({
             'id': data['image_id'],
             'path': 'images' + os.path.sep + data['image_id'] + '.JPG',
             'identity': data['turtle_id'],
-            'position': data['image_location'].str.lower(),
+            'orientation': data['image_location'].str.lower(),
             'split': data['split'],
         })
         return self.finalize_catalogue(df)
\ No newline at end of file
