@@ -1,5 +1,6 @@
 import os
 import argparse
+import shutil
 if __name__ == '__main__':
     import utils
 else:
@@ -7,6 +8,7 @@ else:
 
 def get_data(root):
     with utils.data_directory(root):
+        # TODO: why not the automatic download?
         # Try automatic download
         #import gdown
         #url = 'https://drive.google.com/uc?id=1YT4w8yF44D-y9kdzgF38z2uYbHfpiDOA'
@@ -20,6 +22,14 @@ def get_data(root):
         # Upload to kaggle and download
         os.system(f"kaggle datasets download -d 'vojtacermak/birds' --unzip")
 
+        # Create new folder for segmented images
+        folder_new = os.getcwd() + 'Segmented'
+        if not os.path.exists(folder_new):
+            os.makedirs(folder_new)
+
+        # Move segmented images to new folder
+        folder_move = 'Cropped_pictures'
+        shutil.move(folder_move, os.path.join(folder_new, folder_move))
 
 if __name__ == '__main__':
     parser = argparse.ArgumentParser()
